@@ -7,6 +7,7 @@ import h5py
 
 #turbESN
 from .core import (ESN, _DTYPE, _DEVICE, _ESN_MODES, _WEIGTH_GENERATION, _EXTENDED_STATE_STYLES, _LOGGING_FORMAT)
+from .cross_validation import CrossValidation
 
 #misc
 import sys
@@ -31,7 +32,8 @@ def PreparePredictorData(data: Union[np.ndarray, torch.Tensor],
                          trainingLength: int, 
                          testingLength: int, 
                          esn_start: int, 
-                         esn_end: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
+                         esn_end: int,
+                         validationLength: int=None) -> Union[Tuple[torch.Tensor,torch.Tensor,torch.Tensor,torch.Tensor],Tuple[torch.Tensor,torch.Tensor,torch.Tensor,torch.Tensor,torch.Tensor,torch.Tensor]]:
     ''' Prepares the input and output training and testing/validation data set for the autonomous predictor ESN case, i.e. where in the
         testing pase, the output data is fed back into the reservoir input layer.
         The data set u_test is therefore not needed, as there is only one initial input in the testing phase.
@@ -57,9 +59,17 @@ def PreparePredictorData(data: Union[np.ndarray, torch.Tensor],
     y_train = data_esn[1:trainingLength+1,:]
     
     u_test = data_esn[trainingLength:trainingLength+testingLength,:]
-    y_test = data_esn[trainingLength+1:,:]
+    y_test = data_esn[trainingLength+1:trainingLength+testingLength+1,:]
 
-    return torch.as_tensor(u_train, dtype = _DTYPE), torch.as_tensor(y_train, dtype = _DTYPE),torch.as_tensor(u_test, dtype = _DTYPE),torch.as_tensor(y_test, dtype = _DTYPE)
+    if validationLength is not None:
+        u_val = data_esn[trainingLength+testingLength:trainingLength+testingLength+validationLength,:]
+        y_val = data_esn[trainingLength+testingLength+1:,:]
+    else:
+        u_val = None
+        y_val = None
+
+    return torch.as_tensor(u_train, dtype = _DTYPE), torch.as_tensor(y_train, dtype = _DTYPE),torch.as_tensor(u_test, dtype = _DTYPE),torch.as_tensor(y_test, dtype = _DTYPE),torch.as_tensor(u_val, dtype = _DTYPE),torch.as_tensor(y_val, dtype = _DTYPE)
+    
 #--------------------------------------------------------------------------
 def PrepareTeacherData(data_in: Union[np.ndarray, torch.Tensor], 
                        data_out: int, 
@@ -160,42 +170,26 @@ def InitStudyOrder(nstudy: int, study_parameters: tuple) -> list:
 
     return config
 
-###########################################################################################################
-
-#                            RUNNING AN ESN
-
-###########################################################################################################
-
-def RunturbESN(esn, u_train: Union[np.ndarray, torch.Tensor] = None, 
+#--------------------------------------------------------------------------
+#FH 30.03.2022: added check_user_input (prev. in RunturbESN)
+def check_user_input(esn, u_train: Union[np.ndarray, torch.Tensor] = None, 
                     y_train: Union[np.ndarray, torch.Tensor] = None, 
                     y_test: Union[np.ndarray, torch.Tensor] = None, 
                     pred_init_input: Union[np.ndarray, torch.Tensor] = None, 
                     u_test: Union[np.ndarray, torch.Tensor] = None, 
-                    index_list_auto: list = [], index_list_teacher: list = []) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
-    ''' Runs the turb Echo State Network (turbESN) with the specified parameters. 
-        The training and validation/testing data must be specfied in the argument or in the esn object.
-    
-        INPUT: 
-            esn             - ESN class object containing the reservoir parameters and training & validation/testing data sets.
-            u_train         - training input 
-            y_train         - training output
-            u_test          - testing input
-            y_test          - testing output
-            pred_init_input - initial input in testing phase from which prediction start
+                    u_val: Union[np.ndarray, torch.Tensor] = None, 
+                    y_val: Union[np.ndarray, torch.Tensor] = None,
+                    u_pre_val: Union[np.ndarray, torch.Tensor]=None):
 
-        RETURN:
-            mse_train - mean square error of (teacher forced!) reservoir output (to training target data set y_train) in the training phase. Mean w.r.t. timestep- & mode-axis. 
-            mse_test  - mean square error of reservoir output (to validation data set y_test). Mean w.r.t. timestep- & mode-axis. 
-            y_pred    - reseroir outputs, produced by the given reservoir specified in esn.
-    '''
+    '''Checks data provided by user. The ESN object might be modified by this method.'''
 
-    torch.manual_seed(esn.randomSeed)  
     logging.basicConfig(format=_LOGGING_FORMAT, level= esn.logging_level)
     
     #---------------------------------------
     # Check whether user data is compatiable 
     # (data from esn object is assumed to be correct)
     #---------------------------------------
+
     assert None not in [u_train, y_train, y_test] or None not in [esn.u_train, esn.y_train, esn.y_test], "Error: u_train, y_train or y_test not specified."
 
     if None not in [u_train, y_train, y_test]:
@@ -226,46 +220,211 @@ def RunturbESN(esn, u_train: Union[np.ndarray, torch.Tensor] = None,
         esn.SetTrainingData(u_train=u_train, y_train=y_train)
         esn.SetTestingData(y_test=y_test, pred_init_input=pred_init_input, u_test=u_test)
 
-    esn.toTorch()
-    
-    #---------------------------------------
-    # ESN
-    #---------------------------------------
+    if None not in [u_val, y_val]:
+        if u_pre_val is not None:
+            val_init_input = u_pre_val[-1,:].reshape(1,esn.n_input)
+        else:
+            val_init_input = None
 
+        esn.SetValidationData(y_val=y_val, u_val=u_val, val_init_input=val_init_input)
+
+
+    esn.toTorch()
+
+###########################################################################################################
+
+#                            RUNNING AN ESN
+
+###########################################################################################################
+
+def RunturbESN(esn, u_train: Union[np.ndarray, torch.Tensor]=None, 
+                    y_train: Union[np.ndarray, torch.Tensor]=None, 
+                    y_test: Union[np.ndarray, torch.Tensor]=None, 
+                    pred_init_input: Union[np.ndarray, torch.Tensor]=None, 
+                    u_test: Union[np.ndarray, torch.Tensor]=None, 
+                    index_list_auto: list = [], index_list_teacher: list = [],
+                    u_val: Union[np.ndarray, torch.Tensor]=None,
+                    y_val: Union[np.ndarray, torch.Tensor]=None,
+                    u_pre_val: Union[np.ndarray, torch.Tensor]=None) -> Tuple[torch.Tensor,torch.Tensor,torch.Tensor,torch.Tensor]:
+
+    ''' Runs the turbulent Echo State Network (turbESN) with the specified parameters. 
+        The training and testing data must be specfied in the argument or in the esn object.
+        Optional: validation data (third dataset for cross validation, note naming convention (commonly: training/validation/testing, here: training/testing/validation))
+    
+        INPUT: 
+            esn                - ESN class object. Contains the reservoir parameters. May contain training, testing and validation data sets.
+            u_train            - training input (optional if esn.u_train not None)
+            y_train            - training output (optional if esn.y_train not None)
+            u_test             - testing input (optional if esn.u_test not None)
+            y_test             - testing output (optional if esn.y_test not None)
+            pred_init_input    - initial input in testing phase from which prediction start
+            index_list_auto    - indices of the modes which are passed back as new input (only if esn.mode = _ESN_MODES[2])
+            index_list_teacher - indices of the modes which are supplied by a teacher signal (only if esn.mode = _ESN_MODES[2])
+            u_val              - validation input (optional)
+            y_val              - validation output (optional)
+            u_pre_val          - transient input before validation phase (if None, last reservoir state of testing phase used for init. validation state)
+
+        RETURN:
+            mse_train - mean square error of (teacher forced!) reservoir output (to training target data set y_train) in the training phase. Mean w.r.t. timestep-axis. 
+            mse_test  - mean square error of reservoir output (to testing data set y_test). Mean w.r.t. timestep-axis
+            y_pred    - reseroir outputs, produced by the given reservoir specified in esn.
+            mse_val   - mean square error of reservoir output (to validation data set y_val). Mean w.r.t. timestep-axis. 
+    '''
+
+    torch.manual_seed(esn.randomSeed)  
+    logging.basicConfig(format=_LOGGING_FORMAT, level= esn.logging_level)
+
+    check_user_input(esn,u_train,y_train,y_test,pred_init_input,u_test,u_val,y_val)
+    
+    ##############################################
+    # ESN
+    ##############################################
+    
+    #----------------------------------------------------------
     #1. Create Random Matrices
+    #----------------------------------------------------------
     esn.createWeightMatrices()
 
-    #2. Training Phase 
+    #----------------------------------------------------------
+    #2. Training Phase
+    #----------------------------------------------------------
     esn.x_fit = esn.propagate(u = esn.u_train, transientTime = esn.transientTime)
     esn.fit(X  = esn.x_fit, y = esn.y_train[esn.transientTime:])
-    
     mse_train = ComputeMSE(y_test = esn.y_train[esn.transientTime:], y_pred =  (esn.Wout@esn.x_fit).T)
 
     if np.isnan(esn.Wout).any() or mse_train is None:
         logging.error("Reservoir {0}: while fitting the model, an error occured. Assuming default values.".format(esn.id))
         mse_train = _MSE_DEFAULT
         mse_test = _MSE_DEFAULT
-        y_pred = torch.zeros([esn.testingLength, esn.n_output], device = esn.device, dtype = _DTYPE)
+        mse_val = _MSE_DEFAULT
+        y_pred_test = torch.zeros([esn.testingLength, esn.n_output], device = esn.device, dtype = _DTYPE)
+        y_pred_val = torch.zeros([esn.testingLength, esn.n_output], device = esn.device, dtype = _DTYPE)
 
-        return mse_train, mse_test, y_pred
+        return mse_train, mse_test, mse_val, y_pred_test, y_pred_val
   
-    #3. Prediction Phase (Default:'auto')
+    #----------------------------------------------------------
+    #3. Prediction/Testing Phase (Default:'auto')
+    #----------------------------------------------------------
     if esn.mode == _ESN_MODES[1]:
-        y_pred, esn.x_pred = esn.teacherforce(X = esn.x_fit, testingLength = esn.testingLength)
+        y_pred_test, esn.x_pred = esn.teacherforce(X = esn.x_fit, testingLength = esn.testingLength)
   
     elif esn.mode == _ESN_MODES[2]:
         #TO DO: for now semi-teacher is only possible for n_input = n_output
         assert len(index_list_auto) + len(index_list_teacher) == esn.n_input,'index_list_auto and index_list_teacher do not add up to n_input.'
-        y_pred, esn.x_pred = esn.semiteacherforce(X = esn.x_fit, testingLength = esn.testingLength, 
+        y_pred_test, esn.x_pred = esn.semiteacherforce(X = esn.x_fit, testingLength = esn.testingLength, 
                                                   index_list_auto = index_list_auto, index_list_teacher = index_list_teacher, 
                                                   u_test = esn.u_test)
   
     else:
-        y_pred, esn.x_pred = esn.predict(X = esn.x_fit, testingLength = esn.testingLength)
+        y_pred_test, esn.x_pred = esn.predict(X = esn.x_fit, testingLength = esn.testingLength)
 
-    mse_test = ComputeMSE(y_test = esn.y_test,y_pred = y_pred)
+    mse_test = ComputeMSE(y_test = esn.y_test,y_pred = y_pred_test)
     
-    return mse_train, mse_test, y_pred
+    #----------------------------------------------------------
+    #4. (optional) Validation Phase (for now only in auto mode)
+    #----------------------------------------------------------
+    mse_val = None
+
+    if esn.mode == _ESN_MODES[0]:
+        if None not in [esn.u_val, esn.y_val]:
+
+            validationLength = esn.y_val.shape[0]
+            if u_pre_val is not None:
+                X_pre_val = esn.propagate(u = u_pre_val, transientTime = u_pre_val.shape[0]-1)
+                #esn.val_init_input = u_pre_val[-1,:].reshape(1,esn.n_input)   #this is already set in check_user_input
+            else:
+                # validation phase directly preceeded by testing phase
+                X_pre_val = esn.x_pred
+                esn.val_init_input = esn.y_test[-1,:].reshape(1,esn.n_input)
+
+            y_pred_val, esn.x_val = esn.predict(X=X_pre_val, testingLength=validationLength,init_input=esn.val_init_input)
+            mse_val = ComputeMSE(y_test = esn.y_val,y_pred = y_pred_val)
+        
+
+    return mse_train, mse_test, mse_val, y_pred_test, y_pred_val
+
+#--------------------------------------------------------------------------
+#FH 30.03.2022: added forward_validate_auto_ESN 
+def forward_validate_auto_ESN(esn: ESN, cv: CrossValidation):
+    ''' Runs ESN with k-fold forward walk validation scheme described in Lukosevicius et al. (2021)
+    
+    INPUT:
+        esn    - ESN class object. Contains the reservoir parameters. May contain training, testing and validation data sets.
+        cv     - CrossValidation class object  containing cross validation parameters and data
+        
+    RETURN:
+        MSE_train - mean square error of reservoir output (to training data set y_train) for each fold. Mean w.r.t. mode- & timestep-axis. 
+        MSE_test - mean square error of reservoir output (to training data set y_test) for each fold. Mean w.r.t. mode- & timestep-axis. 
+        MSE_val - mean square error of reservoir output (to training data set y_val) for each fold. Mean w.r.t. mode- & timestep-axis. 
+        Y_PRED_TEST - reseroir outputs of the testing phase for each fold
+        Y_PRED_VAL - reseroir outputs of the validation phase for each fold
+        '''
+
+    max_folds = int(cv.n_folds - cv.n_training_folds - cv.n_validation_folds)
+    transientTime0 = esn.transientTime
+    
+    MSE_train = torch.empty((max_folds,esn.n_output),dtype=_DTYPE)
+    MSE_test = torch.empty((max_folds,esn.n_output),dtype=_DTYPE)
+    MSE_val = torch.empty((max_folds,esn.n_output),dtype=_DTYPE)
+    
+    Y_PRED_TEST = torch.empty((max_folds,esn.testingLength,esn.n_output), dtype=_DTYPE)
+    Y_PRED_VAL  = torch.empty((max_folds,esn.validationLength,esn.n_output), dtype=_DTYPE)
+
+    for ifold in range(max_folds):
+
+        #----------------------------
+        # indices of folds
+        #----------------------------
+        itrain_end = cv.n_training_folds+ifold
+        itest_end = itrain_end + cv.n_testing_folds
+        
+        #----------------------------
+        # length of folds
+        #----------------------------
+        sliding_trainingLength = int(cv.n_training_folds*cv.fold_length) + ifold*cv.fold_length
+        sliding_transientTime = transientTime0 + ifold*cv.fold_length
+        
+        esn.SetTransientTime(sliding_transientTime)
+        esn.trainingLength = sliding_trainingLength
+        
+        #----------------------------
+        # create input/output data
+        #----------------------------
+        u_train = cv.data_folded_u[0:itrain_end].reshape(-1,esn.n_input)
+        y_train = cv.data_folded_y[0:itrain_end].reshape(-1,esn.n_input)
+        
+        u_test = cv.data_folded_u[itrain_end:itest_end].reshape(-1,esn.n_input)
+        y_test = cv.data_folded_y[itrain_end:itest_end].reshape(-1,esn.n_input)
+
+        u_val = cv.data_folded_u[-1].reshape(-1,esn.n_input)
+        y_val = cv.data_folded_y[-1].reshape(-1,esn.n_input)
+
+        if ifold != max_folds -1:
+            u_pre_val = cv.data_folded_u[itest_end:-1].reshape(-1,esn.n_input)
+        else:
+            u_pre_val = None  # validation data is directly preceeded by testing dataset
+
+        #to do: implement more efficient k fold forward walk (precompute esn.x_train, efficient matrix inverse)
+
+        #----------------------------
+        #Run ESN
+        #----------------------------
+        mse_train, mse_test, mse_val, y_pred_test, y_pred_val = RunturbESN(esn = esn, 
+                                                                            u_train = u_train, 
+                                                                            y_train = y_train, 
+                                                                            y_test = y_test, 
+                                                                            u_test = u_test, 
+                                                                            u_val=u_val, 
+                                                                            y_val=y_val,
+                                                                            u_pre_val=u_pre_val)
+       
+        MSE_train[ifold]   = mse_train
+        MSE_test[ifold]    = mse_test
+        MSE_val[ifold]     = mse_val
+        Y_PRED_TEST[ifold] = y_pred_test
+        Y_PRED_VAL[ifold]  = y_pred_val
+        
+    return MSE_train, MSE_test, MSE_val, Y_PRED_TEST, Y_PRED_VAL
 
 ###########################################################################################################
 
@@ -370,29 +529,36 @@ def CreateHDF5Groups(filepath: str, esn_ids: list, nstudy: int):
     #- ...
         
     with h5py.File(filepath, 'a') as f:
-        for esn_id in esn_ids:
-            G = f.create_group(str(esn_id))
+        if type(esn_ids) == int:
+            G = f.create_group(str(esn_ids))
             for studyID in range(nstudy):
-                G.create_group(str(studyID))
+                G.create_group(str(studyID)) 
+        else:
+            for esn_id in esn_ids:
+                G = f.create_group(str(esn_id))
+                for studyID in range(nstudy):
+                    G.create_group(str(studyID))
 
 #--------------------------------------------------------------------------
 def SaveStudy(filepath: str, esn_id: int, 
               studyID: int, study_dict: dict, 
-              y_pred: torch.Tensor, mse_train: torch.Tensor, 
-              mse_test: torch.Tensor, f =None, wass_dist = None):
+              y_pred_test: torch.Tensor, y_pred_val: torch.Tensor,mse_train: torch.Tensor, 
+              mse_test: torch.Tensor, mse_val: torch.Tensor, f =None, wass_dist = None):
     '''Saves the ESN parameters from esn_params into a hdf5 file.
        The h5py file has to be init. with ReadModel (saving the fix parameters) before calling this function!
        
        INPUT:
-          filepath   - path to which the hdf5 file of the ESN study is saved to
-          esn_id     - ESN ID (usually corresponds to a RNG seed)
-          studyID    - ID specifying the study
-          study_dict - dictionary specifying the studied parameter configuration
-          y_pred     - reseroir outputs
-          mse_train  - mean square error of (teacher forced training) reservoir output (to training target dataset y_train) in the training phase. Mean w.r.t. timestep-axis. 
-          mse_test   - mean square error of reservoir output (to validation data set y_test). Mean w.r.t. timestep-axis. 
-          f          - opened hdf5.File where study will be saved to. If None, hdf5 file in filepath will be opened
-          wass_dist  - wasserstein distance of testing data (optional)
+          filepath    - path to which the hdf5 file of the ESN study is saved to
+          esn_id      - ESN ID (usually corresponds to a RNG seed)
+          studyID     - ID specifying the study
+          study_dict  - dictionary specifying the studied parameter configuration
+          y_pred_test - reseroir outputs of testing phase
+          y_pred_val  - reseroir outputs of validation phase
+          mse_train   - mean square error of (teacher forced training) reservoir output (to training target dataset y_train) in the training phase. Mean w.r.t. timestep-axis. 
+          mse_test    - mean square error of reservoir output (to test data set y_test). Mean w.r.t. timestep-axis. 
+          mse_val    - mean square error of reservoir output (to validation data set y_test). Mean w.r.t. timestep-axis. 
+          f           - opened hdf5.File where study will be saved to. If None, hdf5 file in filepath will be opened
+          wass_dist   - wasserstein distance of testing data (optional)
        '''
            
     #HDF5-Structure:
@@ -439,9 +605,12 @@ def SaveStudy(filepath: str, esn_id: int,
     #----------------------------------
     #  Datasets
     #----------------------------------    
-    G_study.create_dataset('y_pred', data = y_pred, compression = 'gzip', compression_opts = 9)
+    G_study.create_dataset('y_pred_test', data = y_pred_test, compression = 'gzip', compression_opts = 9)
+    G_study.create_dataset('y_pred_val', data = y_pred_val, compression = 'gzip', compression_opts = 9)
     G_study.create_dataset('mse_train', data = mse_train, compression = 'gzip', compression_opts = 9)    
     G_study.create_dataset('mse_test', data = mse_test, compression = 'gzip', compression_opts = 9)
+    G_study.create_dataset('mse_val', data = mse_val, compression = 'gzip', compression_opts = 9)
+    
 
     if wass_dist is not None:
         G_study.create_dataset('wass_dist', data = wass_dist)
@@ -491,8 +660,10 @@ def ReadStudy(filepath: str, study_parameters: Union[list,tuple], nstudy: int =
             
         RETURN:
             mse_train   - mean square training error of reservoir output (to validation data set y_train). Mean w.r.t. timestep-axis. 
-            mse_test    - mean square error of reservoir output (to validation data set y_test). Mean w.r.t. timestep-axis. 
-            y_pred      - reseroir outputs, for each study parameter setting of the study 
+            mse_test    - mean square error of reservoir output (to testing data set y_test). Mean w.r.t. timestep-axis. 
+            mse_val     - mean square error of reservoir output (to validation data set y_test). Mean w.r.t. timestep-axis. 
+            y_pred_test - reseroir outputs of testing phase, for each study parameter setting of the study 
+            y_pred_val  - reseroir outputs of validation phase, for each study parameter setting of the study            
             config      - array indicating the parameter setting for given study
     '''
 
@@ -506,7 +677,7 @@ def ReadStudy(filepath: str, study_parameters: Union[list,tuple], nstudy: int =
     if esn_id is not None:  
         with h5py.File(filepath,'r') as f:
 
-            mse_train, mse_test, y_pred, study_dicts = [], [], [], []
+            mse_train, mse_test, mse_val, y_pred_test, y_pred_val, study_dicts = [], [], [], [], [], []
             G_esn_id = f.get(str(esn_id))
             
             #if user does not specify study number, the number is inferred from no. subgroups
@@ -518,10 +689,14 @@ def ReadStudy(filepath: str, study_parameters: Union[list,tuple], nstudy: int =
                 G_study = G_esn_id.get(str(istudy))
 
                 mse_train.append(np.array(G_study.get('mse_train')))
-                mse_train.append(np.array(G_study.get('mse_test')))
+                mse_test.append(np.array(G_study.get('mse_test')))
+                mse_val.append(np.array(G_study.get('mse_val')))
+                
 
                 if read_pred:
-                    y_pred.append(np.array(G_study.get('y_pred')))
+                    y_pred_test.append(np.array(G_study.get('y_pred_test')))
+                    y_pred_val.append(np.array(G_study.get('y_pred_val')))
+                    
 
                 for name in study_parameters:
                     study_dict[name] = G_study.attrs[name]
@@ -537,15 +712,14 @@ def ReadStudy(filepath: str, study_parameters: Union[list,tuple], nstudy: int =
             
             mse_train, mse_test, y_pred, study_dicts = [], [], [], []
 
-            ii = 0
-            for esn_id in esn_ids:
+            for ii,esn_id in enumerate(esn_ids):
                 G_esn_id = f.get(str(esn_id))
             
                 #if user does not specify study number, the number is inferred from no. subgroups
                 if nstudy is None:
                     nstudy = len(G_esn_id.keys())
             
-                MSE_train, MSE_test, Y_pred = [], [], []
+                MSE_train, MSE_test, MSE_val, Y_pred_test, Y_pred_val = [], [], [], [], []
                 for istudy in range(nstudy):
                     G_study = G_esn_id.get(str(istudy))
 
@@ -557,19 +731,24 @@ def ReadStudy(filepath: str, study_parameters: Union[list,tuple], nstudy: int =
             
                     MSE_train.append(np.array(G_study.get('mse_train')))
                     MSE_test.append(np.array(G_study.get('mse_test')))
+                    MSE_val.append(np.array(G_study.get('mse_val')))
 
 
                     if read_pred:
-                        Y_pred.append(np.array(G_study.get('y_pred')))
+                        Y_pred_test.append(np.array(G_study.get('y_pred_test')))
+                        Y_pred_val.append(np.array(G_study.get('y_pred_val')))
+                        
 
                 mse_train.append(MSE_train)
                 mse_test.append(MSE_test)
-                y_pred.append(Y_pred)
+                mse_val.append(MSE_val)
+                
+                y_pred_test.append(Y_pred_test)
+                y_pred_val.append(Y_pred_val)
                 
-                ii+=1
                     
         config = CreateStudyConfigArray(study_parameters, study_dicts)
-        return np.array(mse_train), np.array(mse_test), np.array(y_pred), config
+        return np.array(mse_train), np.array(mse_test), np.array(mse_val),np.array(y_pred_test), np.array(y_pred_val), config
 
 
 ###########################################################################################################
